@@ -3,7 +3,7 @@
   io.c -
 
   $Author: matz $
-  $Date: 1999/04/09 17:57:39 $
+  $Date: 1999/04/26 09:42:31 $
   created at: Fri Oct 15 18:08:59 JST 1993
 
   Copyright (C) 1993-1999 Yukihiro Matsumoto
@@ -702,7 +702,7 @@ lineno_setter(val, id, var)
 }
 
 static VALUE
-arg_set_lineno(argf, val)
+argf_set_lineno(argf, val)
     VALUE argf, val;
 {
     gets_lineno = NUM2INT(val);
@@ -710,7 +710,7 @@ arg_set_lineno(argf, val)
 }
 
 static VALUE
-arg_lineno()
+argf_lineno()
 {
     return lineno;
 }
@@ -1212,13 +1212,14 @@ rb_fdopen(fd, mode)
     return file;
 }
 
-VALUE
-rb_file_open(fname, mode)
+static VALUE
+rb_file_open_internal(klass, fname, mode)
+    VALUE klass;
     const char *fname, *mode;
 {
     OpenFile *fptr;
     NEWOBJ(port, struct RFile);
-    OBJSETUP(port, rb_cFile, T_FILE);
+    OBJSETUP(port, klass, T_FILE);
     MakeOpenFile(port, fptr);
 
     fptr->mode = rb_io_mode_flags(mode);
@@ -1230,7 +1231,15 @@ rb_file_open(fname, mode)
 }
 
 VALUE
-rb_file_sysopen(fname, flags, mode)
+rb_file_open(fname, mode)
+    const char *fname, *mode;
+{
+    return rb_file_open_internal(rb_cFile, fname, mode);
+}
+
+VALUE
+rb_file_sysopen_internal(klass, fname, flags, mode)
+    VALUE klass;
     char *fname;
     int flags, mode;
 {
@@ -1238,13 +1247,13 @@ rb_file_sysopen(fname, flags, mode)
     if (mode != 0666) {
 	rb_warn("can't specify file mode on this platform");
     }
-    return rb_file_open(fname, rb_io_flags_mode(flags));
+    return rb_file_open_internal(klass, fname, rb_io_flags_mode(flags));
 #else
     OpenFile *fptr;
     int fd;
     char *m;
     NEWOBJ(port, struct RFile);
-    OBJSETUP(port, rb_cFile, T_FILE);
+    OBJSETUP(port, klass, T_FILE);
     MakeOpenFile(port, fptr);
 
     fd = rb_open(fname, flags, mode);
@@ -1258,6 +1267,14 @@ rb_file_sysopen(fname, flags, mode)
 #endif
 }
 
+VALUE
+rb_file_sysopen(fname, flags, mode)
+    char *fname;
+    int flags, mode;
+{
+    return rb_file_sysopen_internal(rb_cFile, flags, mode);
+}
+
 #if defined (NT) || defined(DJGPP) || defined(__CYGWIN32__) || defined(__human68k__)
 static struct pipe_list {
     OpenFile *fptr;
@@ -1501,7 +1518,7 @@ rb_file_s_open(argc, argv, klass)
 	int flags = FIX2INT(vmode);
 	int fmode = NIL_P(perm) ? 0666 : FIX2INT(perm);
 
-	file = rb_file_sysopen(path, flags, fmode);
+	file = rb_file_sysopen_internal(klass, path, flags, fmode);
     }
     else {
 	if (!NIL_P(vmode)) {
@@ -1510,11 +1527,9 @@ rb_file_s_open(argc, argv, klass)
 	else {
 	    mode = "r";
 	}
-	file = rb_file_open(RSTRING(fname)->ptr, mode);
+	file = rb_file_open_internal(klass, RSTRING(fname)->ptr, mode);
     }
 
-    RBASIC(file)->klass = klass;
-    rb_obj_call_init(file, 0, 0);
     if (rb_iterator_p()) {
 	return rb_ensure(rb_yield, file, rb_io_close, file);
     }
@@ -1553,6 +1568,7 @@ rb_f_open(argc, argv)
     }
 
     port = pipe_open(RSTRING(pname)->ptr+1, mode);
+    if (NIL_P(port)) return Qnil;
     if (rb_iterator_p()) {
 	return rb_ensure(rb_yield, port, rb_io_close, port);
     }
@@ -1982,7 +1998,7 @@ rb_io_defset(val, id)
 	val = rb_io_open(RSTRING(val)->ptr, "w");
     }
     if (!rb_respond_to(val, id_write)) {
-	rb_raise(rb_eTypeError, "$< must have write method, %s given",
+	rb_raise(rb_eTypeError, "$> must have write method, %s given",
 		 rb_class2name(CLASS_OF(val)));
     }
     rb_defout = val;
@@ -2049,6 +2065,16 @@ rb_io_s_new(argc, argv, klass)
     return prep_stdio(rb_fdopen(NUM2INT(fnum), m), rb_io_mode_flags(m), klass);
 }
 
+static int binmode = 0;
+
+static VALUE
+argf_binmode()
+{
+    rb_io_binmode(file);
+    binmode = 1;
+    return argf;
+}
+
 static int
 next_argv()
 {
@@ -2137,6 +2163,7 @@ next_argv()
 		}
 		file = prep_stdio(fr, FMODE_READABLE, rb_cFile);
 	    }
+	    rb_io_binmode(file);
 	}
 	else {
 	    init_p = 0;
@@ -2265,6 +2292,7 @@ rb_f_backquote(obj, str)
 
     Check_SafeStr(str);
     port = pipe_open(RSTRING(str)->ptr, "r");
+    if (NIL_P(port)) return rb_str_new(0,0);
     result = read_all(port);
 
     rb_io_close(port);
@@ -2687,6 +2715,7 @@ rb_io_s_foreach(argc, argv, io)
 
     arg.argc = argc - 1;
     arg.io = rb_io_open(RSTRING(fname)->ptr, "r");
+    if (NIL_P(arg.io)) return Qnil;
     return rb_ensure(rb_io_foreach_line, (VALUE)&arg, rb_io_close, arg.io);
 }
 
@@ -2718,17 +2747,18 @@ rb_io_s_readlines(argc, argv, io)
 
     arg.argc = argc - 1;
     arg.io = rb_io_open(RSTRING(fname)->ptr, "r");
+    if (NIL_P(arg.io)) return Qnil;
     return rb_ensure(rb_io_readline_line, (VALUE)&arg, rb_io_close, arg.io);
 }
 
 static VALUE
-arg_tell()
+argf_tell()
 {
     return rb_io_tell(file);
 }
 
 static VALUE
-arg_seek(self, offset, ptrname)
+argf_seek(self, offset, ptrname)
      VALUE self, offset, ptrname;
 {
     if (!next_argv()) {
@@ -2739,7 +2769,7 @@ arg_seek(self, offset, ptrname)
 }
 
 static VALUE
-arg_set_pos(self, offset)
+argf_set_pos(self, offset)
      VALUE self, offset;
 {
     if (!next_argv()) {
@@ -2750,25 +2780,25 @@ arg_set_pos(self, offset)
 }
 
 static VALUE
-arg_rewind()
+argf_rewind()
 {
     return rb_io_rewind(file);
 }
 
 static VALUE
-arg_fileno()
+argf_fileno()
 {
     return rb_io_fileno(file);
 }
 
 static VALUE
-arg_to_io()
+argf_to_io()
 {
     return file;
 }
 
 static VALUE
-arg_read(argc, argv)
+argf_read(argc, argv)
     int argc;
     VALUE *argv;
 {
@@ -2802,7 +2832,7 @@ arg_read(argc, argv)
 }
 
 static VALUE
-arg_getc()
+argf_getc()
 {
     VALUE byte;
 
@@ -2819,7 +2849,7 @@ arg_getc()
 }
 
 static VALUE
-arg_readchar()
+argf_readchar()
 {
     VALUE c = rb_io_getc(file);
 
@@ -2830,7 +2860,7 @@ arg_readchar()
 }
 
 static VALUE
-arg_eof()
+argf_eof()
 {
     if (init_p == 0 && !next_argv())
 	return Qtrue;
@@ -2845,11 +2875,11 @@ static VALUE
 rb_f_eof()
 {
     rb_warn("eof? is obsolete; use ARGF.eof? instead");
-    return arg_eof();
+    return argf_eof();
 }
 
 static VALUE
-arg_each_line(argc, argv)
+argf_each_line(argc, argv)
     int argc;
     VALUE *argv;
 {
@@ -2862,30 +2892,30 @@ arg_each_line(argc, argv)
 }
 
 static VALUE
-arg_each_byte()
+argf_each_byte()
 {
     VALUE byte;
 
-    while (!NIL_P(byte = arg_getc())) {
+    while (!NIL_P(byte = argf_getc())) {
 	rb_yield(byte);
     }
     return Qnil;
 }
 
 static VALUE
-arg_filename()
+argf_filename()
 {
     return filename;
 }
 
 static VALUE
-arg_file()
+argf_file()
 {
     return file;
 }
 
 static VALUE
-arg_skip()
+argf_skip()
 {
     if (next_p != -1) {
 	rb_io_close(file);
@@ -2895,7 +2925,7 @@ arg_skip()
 }
 
 static VALUE
-arg_close()
+argf_close()
 {
     rb_io_close(file);
     if (next_p != -1) {
@@ -2906,7 +2936,7 @@ arg_close()
 }
 
 static VALUE
-arg_closed()
+argf_closed()
 {
     return rb_io_closed(file);
 }
@@ -3062,37 +3092,38 @@ Init_IO()
     rb_define_readonly_variable("$<", &argf);
     rb_define_global_const("ARGF", argf);
 
-    rb_define_singleton_method(argf, "fileno", arg_fileno, 0);
-    rb_define_singleton_method(argf, "to_i", arg_fileno, 0);
-    rb_define_singleton_method(argf, "to_io", arg_to_io, 0);
-    rb_define_singleton_method(argf, "each",  arg_each_line, -1);
-    rb_define_singleton_method(argf, "each_line",  arg_each_line, -1);
-    rb_define_singleton_method(argf, "each_byte",  arg_each_byte, 0);
+    rb_define_singleton_method(argf, "fileno", argf_fileno, 0);
+    rb_define_singleton_method(argf, "to_i", argf_fileno, 0);
+    rb_define_singleton_method(argf, "to_io", argf_to_io, 0);
+    rb_define_singleton_method(argf, "each",  argf_each_line, -1);
+    rb_define_singleton_method(argf, "each_line",  argf_each_line, -1);
+    rb_define_singleton_method(argf, "each_byte",  argf_each_byte, 0);
 
-    rb_define_singleton_method(argf, "read",  arg_read, -1);
+    rb_define_singleton_method(argf, "read",  argf_read, -1);
     rb_define_singleton_method(argf, "readlines", rb_f_readlines, -1);
     rb_define_singleton_method(argf, "to_a", rb_f_readlines, -1);
     rb_define_singleton_method(argf, "gets", rb_f_gets, -1);
     rb_define_singleton_method(argf, "readline", rb_f_readline, -1);
-    rb_define_singleton_method(argf, "getc", arg_getc, 0);
-    rb_define_singleton_method(argf, "readchar", arg_readchar, 0);
-    rb_define_singleton_method(argf, "tell", arg_tell, 0);
-    rb_define_singleton_method(argf, "seek", arg_seek, 2);
-    rb_define_singleton_method(argf, "rewind", arg_rewind, 0);
-    rb_define_singleton_method(argf, "pos", arg_tell, 0);
-    rb_define_singleton_method(argf, "pos=", arg_set_pos, 1);
-    rb_define_singleton_method(argf, "eof", arg_eof, 0);
-    rb_define_singleton_method(argf, "eof?", arg_eof, 0);
+    rb_define_singleton_method(argf, "getc", argf_getc, 0);
+    rb_define_singleton_method(argf, "readchar", argf_readchar, 0);
+    rb_define_singleton_method(argf, "tell", argf_tell, 0);
+    rb_define_singleton_method(argf, "seek", argf_seek, 2);
+    rb_define_singleton_method(argf, "rewind", argf_rewind, 0);
+    rb_define_singleton_method(argf, "pos", argf_tell, 0);
+    rb_define_singleton_method(argf, "pos=", argf_set_pos, 1);
+    rb_define_singleton_method(argf, "eof", argf_eof, 0);
+    rb_define_singleton_method(argf, "eof?", argf_eof, 0);
+    rb_define_singleton_method(argf, "binmode", argf_binmode, 0);
 
-    rb_define_singleton_method(argf, "to_s", arg_filename, 0);
-    rb_define_singleton_method(argf, "filename", arg_filename, 0);
-    rb_define_singleton_method(argf, "file", arg_file, 0);
-    rb_define_singleton_method(argf, "skip", arg_skip, 0);
-    rb_define_singleton_method(argf, "close", arg_close, 0);
-    rb_define_singleton_method(argf, "closed?", arg_closed, 0);
+    rb_define_singleton_method(argf, "to_s", argf_filename, 0);
+    rb_define_singleton_method(argf, "filename", argf_filename, 0);
+    rb_define_singleton_method(argf, "file", argf_file, 0);
+    rb_define_singleton_method(argf, "skip", argf_skip, 0);
+    rb_define_singleton_method(argf, "close", argf_close, 0);
+    rb_define_singleton_method(argf, "closed?", argf_closed, 0);
 
-    rb_define_singleton_method(argf, "lineno",   arg_lineno, 0);
-    rb_define_singleton_method(argf, "lineno=",  arg_set_lineno, 1);
+    rb_define_singleton_method(argf, "lineno",   argf_lineno, 0);
+    rb_define_singleton_method(argf, "lineno=",  argf_set_lineno, 1);
 
     file = rb_stdin;
     rb_global_variable(&file);
