@@ -37,6 +37,7 @@ $libdir += "/"+CONFIG["MAJOR"]+"."+CONFIG["MINOR"]
 $archdir = $libdir+"/"+CONFIG["arch"]
 $install = CONFIG["INSTALL_PROGRAM"]
 $install_data = CONFIG["INSTALL_DATA"]
+$ruby_inc = $archdir
 if $install =~ %r!^[^\s/]+/! then
   $install = CONFIG["srcdir"]+"/"+$install
   $install_data = CONFIG["srcdir"]+"/"+$install_data
@@ -60,8 +61,8 @@ if /win32|djgpp|mingw32|m68k-human|i386-os2_emx/i =~ PLATFORM
 else
   $null = open("/dev/null", "w")
 end
-LINK = "#{CONFIG['CC']} -o conftest -I#{$srcdir} -I#{CONFIG['includedir']} #{CFLAGS} %s #{CONFIG['LDFLAGS']} %s conftest.c #{CONFIG['LIBS']} %s"
-CPP = "#{CONFIG['CPP']} -E -I#{$srcdir} -I#{CONFIG['includedir']} #{CFLAGS} %s conftest.c"
+LINK = "#{CONFIG['CC']} -o conftest -I#{$hdrdir} -I#{CONFIG['includedir']} #{CFLAGS} %s #{CONFIG['LDFLAGS']} %s conftest.c #{CONFIG['LIBS']} %s"
+CPP = "#{CONFIG['CPP']} -E -I#{$hdrdir} -I#{CONFIG['includedir']} #{CFLAGS} %s conftest.c"
 
 $orgerr = $stderr.dup
 $orgout = $stdout.dup
@@ -78,12 +79,39 @@ def xsystem command
   return r
 end
 
-def try_link(libs)
-  xsystem(format(LINK, $CFLAGS, $LDFLAGS, libs))
+def try_link(src, opt="")
+  cfile = open("conftest.c", "w")
+  cfile.print src
+  cfile.close
+  begin
+    xsystem(format(LINK, $CFLAGS, $LDFLAGS, opt))
+  ensure
+    system "rm -f conftest.*"
+  end
 end
 
-def try_cpp
-  xsystem(format(CPP, $CFLAGS))
+def try_cpp(src, opt=$CFLAGS)
+  cfile = open("conftest.c", "w")
+  cfile.print src
+  cfile.close
+  xsystem(format(CPP, opt))
+end
+
+def egrep_cpp(pat, src, opt=$CFLAGS)
+  cfile = open("conftest.c", "w")
+  cfile.print src
+  cfile.close
+  xsystem(format(CPP+"|egrep #{pat}", opt))
+end
+
+def try_run(src, opt="")
+  begin
+    if try_link(src, opt)
+      xsystem("./conftest")
+    end
+  ensure
+    system "rm -f conftest*"
+  end
 end
 
 def install_rb(mfile)
@@ -136,7 +164,10 @@ int t() { %s(); return 0; }
       else
 	libs = "-l" + lib
       end
-      unless try_link(libs)
+      unless try_link(<<"SRC", libs)
+int main() { return 0; }
+int t() { #{func}(); return 0; }
+SRC
 	$lib_cache[lib] = 'no'
 	$cache_mod = TRUE
 	print "no\n"
@@ -174,19 +205,15 @@ def have_func(func)
     end
   end
 
-  cfile = open("conftest.c", "w")
-  cfile.printf "\
-char %s();
-int main() { return 0; }
-int t() { %s(); return 0; }
-", func, func
-  cfile.close
-
   libs = $libs
   libs = "" if libs == nil
 
   begin
-    unless try_link(libs)
+      unless try_link(<<"SRC", libs)
+char #{func}();
+int main() { return 0; }
+int t() { #{func}(); return 0; }
+SRC
       $func_found[func] = 'no'
       $found = TRUE
       print "no\n"
@@ -217,14 +244,10 @@ def have_header(header)
     end
   end
 
-  cfile = open("conftest.c", "w")
-  cfile.printf "\
-#include <%s>
-", header
-  cfile.close
-
   begin
-    unless try_cpp
+    unless try_cpp(<<"SRC")
+#include <#{header}>
+SRC
       $hdr_found[header] = 'no'
       $found = TRUE
       print "no\n"
@@ -241,6 +264,38 @@ def have_header(header)
   return TRUE
 end
 
+def arg_config(config, default="yes")
+  unless defined? $configure_args
+    $configure_args = {}
+    for arg in CONFIG["configure_args"].split + ARGV
+      next unless /^--/ =~ arg
+      if /=/ =~ arg
+	$configure_args[$`] = $'
+      else
+	$configure_args[arg] = default
+      end
+    end
+  end
+  $configure_args.fetch(config, default)
+end
+
+def with_config(config, default="yes")
+  unless /^--with-/ =~ config
+    config = '--with-' + config
+  end
+  arg_config(config, default)
+end
+
+def enable_config(config, default=nil)
+  if arg_config("--enable-"+config, true)
+    true
+  elsif arg_config("--disable-"+config, false)
+    false
+  else
+    default
+  end
+end
+
 def create_header()
   print "creating extconf.h\n"
   STDOUT.flush
@@ -291,7 +346,6 @@ hdrdir = #{$hdrdir}
 
 CC = #{CONFIG["CC"]}
 
-prefix = #{CONFIG["prefix"]}
 CFLAGS   = #{CONFIG["CCDLFLAGS"]} -I$(hdrdir) -I#{CONFIG["includedir"]} #{CFLAGS} #{$CFLAGS} #{$defs.join(" ")}
 CXXFLAGS = $(CFLAGS)
 DLDFLAGS = #{$DLDFLAGS} #{$LDFLAGS}
@@ -301,6 +355,7 @@ prefix = #{CONFIG["prefix"]}
 exec_prefix = #{CONFIG["exec_prefix"]}
 libdir = #{$libdir}
 archdir = #{$archdir}
+ruby_inc = #{$ruby_inc}
 
 #### End of system configuration section. ####
 
