@@ -3,11 +3,10 @@
   re.h -
 
   $Author: matz $
-  $Revision: 1.1.1.2.2.3 $
-  $Date: 1998/11/26 08:25:46 $
+  $Date: 1999/01/26 10:08:14 $
   created at: Thu Sep 30 14:18:32 JST 1993
 
-  Copyright (C) 1993-1998 Yukihiro Matsumoto
+  Copyright (C) 1993-1999 Yukihiro Matsumoto
 
 ************************************************/
 
@@ -29,8 +28,12 @@ struct RMatch {
 
 #define RMATCH(obj)  (R_CAST(RMatch)(obj))
 
-int str_cicmp _((VALUE, VALUE));
-VALUE reg_regcomp _((VALUE));
-int reg_search _((VALUE, VALUE, int, int));
-VALUE reg_regsub _((VALUE, VALUE, struct re_registers *));
+int rb_str_cicmp _((VALUE, VALUE));
+VALUE rb_reg_regcomp _((VALUE));
+int rb_reg_search _((VALUE, VALUE, int, int));
+VALUE rb_reg_regsub _((VALUE, VALUE, struct re_registers *));
+
+int rb_kcode _((void));
+
+extern int ruby_ignorecase;
 #endif
