@@ -3,7 +3,7 @@
   string.c -
 
   $Author: matz $
-  $Date: 1999/03/24 08:52:33 $
+  $Date: 1999/04/26 09:42:42 $
   created at: Mon Aug  9 17:12:58 JST 1993
 
   Copyright (C) 1993-1999 Yukihiro Matsumoto
@@ -865,8 +865,8 @@ rb_str_aset(str, indx, val)
 	if (idx < 0) {
 	    idx += RSTRING(str)->len;
 	}
-	if (idx < 0 || RSTRING(str)->len < idx) {
-	    rb_raise(rb_eIndexError, "index %d out of string", NUM2INT(beg));
+	if (idx < 0 || RSTRING(str)->len <= idx) {
+	    rb_raise(rb_eIndexError, "index %d out of string", idx);
 	}
 	if (FIXNUM_P(val)) {
 	    if (RSTRING(str)->len == idx) {
@@ -1404,7 +1404,7 @@ rb_str_dump(str)
 	switch (c) {
 	  case '"':  case '\\':
 	  case '\n': case '\r':
-	  case '\t': case '\f':
+	  case '\t': case '\f': case '#':
 	  case '\013': case '\007': case '\033': 
 	    len += 2;
 	    break;
@@ -1432,6 +1432,10 @@ rb_str_dump(str)
 	    *q++ = '\\';
 	    *q++ = c;
 	}
+	else if (c == '#') {
+	    *q++ = '\\';
+	    *q++ = '#';
+	}
 	else if (ISPRINT(c)) {
 	    *q++ = c;
 	}
@@ -2182,7 +2186,7 @@ rb_f_chop()
     VALUE str = rb_str_dup(uscore_get());
     VALUE val = rb_str_chop_bang(str);
 
-    if (NIL_P(str)) return str;
+    if (NIL_P(val)) return str;
     rb_lastline_set(val);
     return val;
 }
@@ -2257,6 +2261,9 @@ rb_f_chomp(argc, argv)
 {
     VALUE str = rb_str_dup(uscore_get());
     VALUE val = rb_str_chomp_bang(argc, argv, str);
+
+    if (NIL_P(val)) return str;
+    rb_lastline_set(val);
     return val;
 }
 
